﻿//--------------------------------------------------------------------------------
// <copyright file="MergeDiffFiles.cs" 
//            company="The University of Queensland"
//            author="Timothy O'Connor">
//     Copyright © The University of Queensland, 2012-2014. All rights reserved.
// </copyright>
// License: 
//--------------------------------------------------------------------------------

namespace Analyses
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using Genomics;
    using Shared;
    using Data;
    using Tools;

    /// <summary>
    /// Merge diff files.
    /// </summary>
    public class MergeDiffFiles
    {
        /// <summary>
        /// The set names.
        /// </summary>
        private List<string> setNames;

        /// <summary>
        /// The diff files.
        /// </summary>
        private List<CuffDiffFile> diffFiles;

        /// <summary>
        /// The link files.
        /// </summary>
        private List<GtfExpressionFile> expressionFiles;

        /// <summary>
        /// The sample expression file.
        /// </summary>
        private Dictionary<string, GtfExpressionFile> sampleExpressionFile;

        /// <summary>
        /// The set gene fpkm selector.
        /// </summary>
        private Dictionary<string, Func<string, double>> setGeneFpkmSelector;

        /// <summary>
        /// The genes.
        /// </summary>
        private List<string> genes;

        /// <summary>
        /// Correlation Maps.
        /// </summary>
        private List<TssRegulatoryMap> nearestMotifMaps;

        /// <summary>
        /// The maps.
        /// </summary>
        private List<TssRegulatoryMap> maps;

        /// <summary>
        /// Null map.
        /// </summary>
        private TssRegulatoryMap nullMap;

        /// <summary>
        /// The gene motif distance.
        /// </summary>
        private Dictionary<string, int> geneMotifDistance;

        /// <summary>
        /// The gene motif distance rank.
        /// </summary>
        private Dictionary<string, int> geneMotifDistanceRank;

        /// <summary>
        /// The gene nearest motif correlation score.
        /// </summary>
        private Dictionary<string, double> geneNearestMotifCorrelationScore;

        /// <summary>
        /// The gene nearest motif correlation rank.
        /// </summary>
        private Dictionary<string, int> geneNearestMotifCorrelationRank;

        /// <summary>
        /// The gene motif correlation score.
        /// </summary>
        private Dictionary<string, double> geneMotifCorrelationScore;

        /// <summary>
        /// The gene motif correlation rank.
        /// </summary>
        private Dictionary<string, int> geneMotifCorrelationRank;
       
        /// <summary>
        /// Initializes a new instance of the <see cref="Analyses.MergeDiffFiles"/> class.
        /// </summary>
        public MergeDiffFiles()
        {
        }

        /// <summary>
        /// Gets or sets the diff file names.
        /// </summary>
        /// <value>The diff file names.</value>
        public string[] DiffFileNames { get; set; }

        /// <summary>
        /// Gets or sets the cufflinks expression file names.
        /// </summary>
        /// <value>The link file names.</value>
        public string[] LinkFileNames { get; set; }

        /// <summary>
        /// Gets or sets the diff definitions.
        /// </summary>
        /// <value>The diff definitions.</value>
        public List<Tuple<string, string>> DiffDefinitions { get; set; }

        /// <summary>
        /// Gets or sets the link set definitions.
        /// </summary>
        /// <value>The link set definitions.</value>
        public string[] LinkDefinitions { get; set; }

        /// <summary>
        /// Gets or sets the map file names.
        /// </summary>
        /// <value>The map file names.</value>
        public string[] MapFileNames { get; set; }

        /// <summary>
        /// Gets or sets the null map file name.
        /// </summary>
        /// <value>The null map file.</value>
        public string NullMapFile { get; set; }

        /// <summary>
        /// Gets or sets the output filen.
        /// </summary>
        /// <value>The output filen.</value>
        public string OutputFile { get; set; }

        /// <summary>
        /// Gets the diff files.
        /// </summary>
        /// <value>The diff files.</value>
        public List<CuffDiffFile> DiffFiles
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.diffFiles,
                    () => this.DiffFileNames
                        .Select(x => new CuffDiffFile(x))
                        .ToList());
            }
        }

        /// <summary>
        /// Gets the expression files.
        /// </summary>
        /// <value>The expression files.</value>
        public List<GtfExpressionFile> ExpressionFiles
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.expressionFiles,
                    () => this.LinkFileNames
                        .Select(x => new GtfExpressionFile(GtfExpressionFile.ExpressionType.LongPap, x))
                        .ToList());
            }
        }

        // TODO: remove infinities
        // TODO: add individual sample along mice
        // TODO: add pvalues 
        // TODO: flip diff direction

        /// <summary>
        /// Gets the set names.
        /// </summary>
        /// <value>The set names.</value>
        public List<string> SetNames
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.setNames,
                    () => this.DiffDefinitions
                        .Select(x => new List<string> { x.Item1, x.Item2 })
                        .SelectMany(x => x)
                        .Distinct()
                        .OrderBy(x => x)
                        .ToList());
            }
        }

        /// <summary>
        /// Maps sample names to their expression files
        /// </summary>
        /// <value>The sample expression file.</value>
        public Dictionary<string, GtfExpressionFile> SampleExpressionFile
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.sampleExpressionFile,
                    () => this.LinkDefinitions
                        .Select((x, i) => new { Sample = x, Index = i })
                        .ToDictionary(x => x.Sample, x => this.ExpressionFiles[x.Index]));
            }
        }

        /// <summary>
        /// Gets the diff names.
        /// </summary>
        /// <value>The diff names.</value>
        public List<string> DiffNames
        {
            get
            {
                return this.DiffDefinitions.Select(x => string.Format("{0}_vs_{1}", x.Item1, x.Item2)).ToList();
            }
        }

        /// <summary>
        /// Gets the set gene fpkm selector.
        /// </summary>
        /// <value>The set gene fpkm selector.</value>
        public Dictionary<string, Func<string, double>> SetGeneFpkmSelector
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.setGeneFpkmSelector,
                    () => this.SetNames.Select(x => 
                    {
                        int i = this.DiffDefinitions.FindIndex(y => y.Item1 == x);
                        Func<string, double> selector;

                        if (i >= 0)
                        {
                            selector = (gene) => this.DiffFiles[i].TranscriptData.ContainsKey(gene) ? this.DiffFiles[i].TranscriptData[gene].Fpkm1 : double.NaN;
                        }
                        else
                        {
                            i = this.DiffDefinitions.FindIndex(y => y.Item2 == x);
                            if (i >= 0)
                            {
                                selector = (gene) => this.DiffFiles[i].TranscriptData.ContainsKey(gene) ? this.DiffFiles[i].TranscriptData[gene].Fpkm2 : double.NaN;
                            }
                            else
                            {
                                throw new Exception("Unknown set " + x);
                            }
                        }

                        return new
                        {
                            Set = x,
                            Selector = selector,
                        };
                    })
                    .ToDictionary(x => x.Set, x => x.Selector));
            }
        }

        /// <summary>
        /// Gets the maps.
        /// </summary>
        /// <value>The maps.</value>
        public List<TssRegulatoryMap> NearestMotifMaps 
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.nearestMotifMaps,
                    () => this.MapFileNames
                        .Select(x => TssRegulatoryMap.LoadMap(
                            x, 
                            new MapLinkFilter 
                            { 
                                LinkTypeFilter = MapLinkFilter.LinkType.NearestCrm 
                            }))
                        .ToList());
            }
        }

        public List<TssRegulatoryMap> Maps 
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.maps,
                    () => this.MapFileNames
                    .Select(x => TssRegulatoryMap.LoadMap(
                        x, 
                        new MapLinkFilter 
                    { 
                        LinkTypeFilter = MapLinkFilter.LinkType.Any 
                    }))
                    .ToList());
            }
        }

        /// <summary>
        /// Gets the null map.
        /// </summary>
        /// <value>The null map.</value>
        public TssRegulatoryMap NullMap
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.nullMap,
                    () => TssRegulatoryMap.LoadMap(
                        this.NullMapFile, 
                        new MapLinkFilter
                        { 
                            LinkTypeFilter = MapLinkFilter.LinkType.NearestCrm 
                        }));
            }
        }

        public Dictionary<string, int> GeneNearestMotifDistance
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.geneMotifDistance,
                    () => this.NullMap.Links.ToDictionary(x => (string)x.TranscriptName, x => x.AbsLinkLength));
            }
        }

        public Dictionary<string, int> GeneNearestMotifDistanceRank
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.geneMotifDistanceRank,
                    () => this.GeneNearestMotifDistance
                        .OrderBy(x => x.Value)
                        .Select((x, i) => new { Gene = x.Key, Rank = i })
                        .ToDictionary(x => x.Gene, x => x.Rank));
            }
        }

        public Dictionary<string, double> GeneNearestMotifCorrelationScore
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.geneNearestMotifCorrelationScore,
                    () => this.NearestMotifMaps.First().Links.ToDictionary(x => (string)x.TranscriptName, x => x.ConfidenceScore));
            }
        }

        public Dictionary<string, int> GeneNearestMotifCorrelationRank
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.geneNearestMotifCorrelationRank,
                    () => 
                {
                    var scaledRanks = this.GeneNearestMotifCorrelationScore
                        .OrderBy(x => x.Value)
                        .Select((x, i) => new { Gene = x.Key, Rank = (double)i / this.GeneNearestMotifCorrelationScore.Count * this.GeneNearestMotifDistanceRank.Count })
                        .ToDictionary(x => x.Gene, x => x.Rank);

                    return this.NullMap.Links
                        .Select(x => new 
                        {
                            Gene = (string)x.TranscriptName,
                            Rank = scaledRanks.ContainsKey(x.TranscriptName) ? scaledRanks[x.TranscriptName] : double.NaN,
                        })
                        .OrderBy(x => x.Rank)
                        .Select((x, i) => new { Gene = x.Gene, Rank = i })
                        .ToDictionary(x => x.Gene, x => x.Rank);
                });
            }
        }

        public Dictionary<string, double> GeneMotifCorrelationScore
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.geneMotifCorrelationScore,
                    () => this.Maps.First().GetBestNeighborMap(1).Links.ToDictionary(x => (string)x.TranscriptName, x => x.ConfidenceScore));
            }
        }

        public Dictionary<string, int> GeneMotifCorrelationRank
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.geneMotifCorrelationRank,
                    () => 
                {
                    var scaledRanks = this.GeneMotifCorrelationScore
                        .OrderBy(x => x.Value)
                        .Select((x, i) => new { Gene = x.Key, Rank = (double)i / this.GeneMotifCorrelationScore.Count * this.GeneNearestMotifDistanceRank.Count })
                        .ToDictionary(x => x.Gene, x => x.Rank);

                    return this.NullMap.Keys
                        .Select(x => new 
                        {
                            Gene = (string)x,
                            Rank = scaledRanks.ContainsKey(x) ? scaledRanks[x] : double.NaN,
                        })
                        .OrderBy(x => x.Rank)
                        .Select((x, i) => new { Gene = x.Gene, Rank = i })
                        .ToDictionary(x => x.Gene, x => x.Rank);
                });
            }
        }

        /// <summary>
        /// Gets the genes.
        /// </summary>
        /// <value>The genes.</value>
        public List<string> Genes
        {
            get
            {
                return Helpers.CheckInit(
                    ref this.genes,
                    () => this.DiffFiles
                        .Select(x => x.TranscriptData.Keys)
                        .SelectMany(x => x)
                        .ToLookup(x => x, x => x)
                        .Select(x => x.Key)
                        .OrderBy(x => x)
                        .ToList());
            }
        }

        /// <summary>
        /// Execute this instance.
        /// </summary>
        public void Execute()
        {
            Tables.ToNamedTsvFile(
                this.OutputFile,
                this.Genes
                    .Select((gene => new string[]
                        {
                            gene,
                        }
                        .Concat(this.GetSetFpkms(gene).Select(x => x.ToString()))
                        .Concat(this.GetLinkFpkms(gene).Select(x => x.ToString()))
                        .Concat(this.GetDiffFoldChanges(gene).Select(x => x.ToString()))
                        .Concat(this.GetDiffPvalues(gene).Select(x => x.ToString()))
                        .Concat(this.GetDiffQvalues(gene).Select(x => x.ToString()))
                        .Concat(new double[] 
                        {
                            this.GeneNearestMotifDistance.ContainsKey(gene) ? this.GeneNearestMotifDistance[gene] : double.NaN,
                            this.GeneNearestMotifDistance.ContainsKey(gene) ? this.GeneNearestMotifDistanceRank[gene] : double.NaN,
                            this.GeneNearestMotifCorrelationScore.ContainsKey(gene) ? this.GeneNearestMotifCorrelationScore[gene] : double.NaN,
                            this.GeneNearestMotifCorrelationRank.ContainsKey(gene) ? this.GeneNearestMotifCorrelationRank[gene] : double.NaN,
                            this.GeneMotifCorrelationScore.ContainsKey(gene) ? this.GeneMotifCorrelationScore[gene] : double.NaN,
                            this.GeneMotifCorrelationRank.ContainsKey(gene) ? this.GeneMotifCorrelationRank[gene] : double.NaN,
                        }
                            .Select(x => x.ToString()))
                        .ToList()))
                    .ToList(),
                new string[] { "Gene" }
                    .Concat(this.SetNames)
                    .Concat(this.LinkDefinitions)
                    .Concat(this.DiffNames)
                    .Concat(this.DiffNames.Select(x => x + "_pval"))
                    .Concat(this.DiffNames.Select(x => x + "_qval"))
                    .Concat(new string[] { "MotifDistance", "DistanceRank", "NearestMotifScore", "NearestMotifRank", "BestMotifScore", "BestMotifRank" })
                    .ToList());
        }

        /// <summary>
        /// Gets the set fpkms.
        /// </summary>
        /// <returns>The set fpkms.</returns>
        /// <param name="gene">Gene.</param>
        protected List<double> GetSetFpkms(string gene)
        {
            return this.SetNames.Select(x => this.SetGeneFpkmSelector[x](gene)).ToList();
        }

        /// <summary>
        /// Gets the set sample fpkms.
        /// </summary>
        /// <returns>The set sample fpkms.</returns>
        /// <param name="gene">Gene.</param>
        protected List<double> GetLinkFpkms(string gene)
        {
            return this.LinkDefinitions
                .Select(x => this.SampleExpressionFile[x].Genes.ContainsKey(gene) ?
                    this.SampleExpressionFile[x].Genes[gene].Score :
                    double.NaN)
                .ToList();
        }

        /// <summary>
        /// Gets the diff fold changes.
        /// </summary>
        /// <returns>The diff fold changes.</returns>
        /// <param name="gene">Gene.</param>
        protected List<double> GetDiffFoldChanges(string gene)
        {
            return this.DiffFiles.Select(x => x.TranscriptData.ContainsKey(gene) ? x.TranscriptData[gene].FoldChange : double.NaN).ToList();
        }

        /// <summary>
        /// Gets the diff pvalues.
        /// </summary>
        /// <returns>The diff pvalues.</returns>
        /// <param name="gene">Gene.</param>
        protected List<double> GetDiffPvalues(string gene)
        {
            return this.DiffFiles.Select(x => x.TranscriptData.ContainsKey(gene) ? x.TranscriptData[gene].PValue : double.NaN).ToList();
        }

        /// <summary>
        /// Gets the diff qvalues.
        /// </summary>
        /// <returns>The diff qvalues.</returns>
        /// <param name="gene">Gene.</param>
        protected List<double> GetDiffQvalues(string gene)
        {
            return this.DiffFiles.Select(x => x.TranscriptData.ContainsKey(gene) ? x.TranscriptData[gene].QValue : double.NaN).ToList();
        }

        /// <summary>
        /// Execute the ChIP seq peak bias analysis.
        /// </summary>
        public class Executor : IAnalysisExecutor<MergeDiffFiles, MergeDiffFiles.Executor.Arguments>
        {
            /// <summary>
            /// The options.
            /// </summary>
            protected override Dictionary<Arguments, string> OptionsData
            {
                get
                {
                    return new Dictionary<Arguments, string>
                    {
                        { Arguments.DiffFileNames, "CSV list of diff files to merge" },
                        { Arguments.DiffDefinitions, "Name-value pair list of sets compared in the diff files" },
                        { Arguments.LinkFileNames, "CSV list of cufflinks files to merge" },
                        { Arguments.LinkDefinitions, "Name list of sets and individual link samples" },
                        { Arguments.MapFileNames, "CSV list of map file names" },
                        { Arguments.NullMapFile, "Null map file name" },

                        { Arguments.OutputFile, "Output file to store the table in" },
                    };
                }
            }

            /// <summary>
            /// Commany line arguments to the analysis
            /// </summary>
            public enum Arguments
            {
                /// <summary>
                /// The diff files in order to be merged
                /// </summary>
                DiffFileNames,

                /// <summary>
                /// Name-value pair list of sets represented by diffs
                /// </summary>
                DiffDefinitions,

                /// <summary>
                /// The cufflinks files of individual samples to merge
                /// </summary>
                LinkFileNames,

                /// <summary>
                /// Name list of the sets and their individual samples
                /// </summary>
                LinkDefinitions,

                /// <summary>
                /// CSV list of map file names
                /// </summary>
                MapFileNames,
              
                /// <summary>
                /// Null map file
                /// </summary>
                NullMapFile,

                /// <summary>
                /// The output file.
                /// </summary>
                OutputFile,
            }

            /// <summary>
            /// Gets the description.
            /// </summary>
            /// <value>The description.</value>
            public override string Description
            {
                get
                {
                    return "Merges the output of multiple differential expression experiments";
                }
            }

            /// <summary>
            /// Execute the analysis
            /// </summary>
            /// <param name = "commandArgs">Command line arguments</param>
            public override void Execute(Args commandArgs)
            {
                var analysis = new MergeDiffFiles();

                analysis.DiffFileNames = commandArgs.StringEnumArgs[Arguments.DiffFileNames].Split(',');
                analysis.DiffDefinitions = commandArgs.StringEnumArgs[Arguments.DiffDefinitions].Split(';')
                    .Select(x => x.Split(','))
                    .Select(x => new Tuple<string, string>(x[0], x[1]))
                    .ToList();
                analysis.LinkFileNames = commandArgs.StringEnumArgs[Arguments.LinkFileNames].Split(',');
                analysis.LinkDefinitions = commandArgs.StringEnumArgs[Arguments.LinkDefinitions].Split(',');
                analysis.OutputFile = commandArgs.StringEnumArgs[Arguments.OutputFile];
                analysis.NullMapFile = commandArgs.StringEnumArgs[Arguments.NullMapFile];
                analysis.MapFileNames = commandArgs.StringEnumArgs[Arguments.MapFileNames].Split(',');

                analysis.Execute();
            }
        }
    }
}

