/// <summary>
        /// Multiples the regression data use xml config.
        /// </summary>
        /// <param name="xmlFile">Xml file.</param>
        /// <param name="dataSource">Data source.</param>
        /// <param name="cellLineOverride">Cell line override.</param>
        /// <param name="generateFeatures">If set to <c>true</c> generate features.</param>
        /// <param name="sampleDrms">If set to <c>true</c> sample drms.</param>
        /// <param name="useFullMap">If set to <c>true</c> use full map.</param>
        /// <param name="startRep">Start rep.</param>
        /// <param name="repCount">Rep count.</param>
        /// <param name="batch">Batch.</param>
        /// <param name="stage">Stage.</param>
        private static void MultipleRegressionDataUseXmlConfig(
            string xmlFile, 
            string dataSource,
            string cellLineOverride,
            bool generateFeatures,
            bool sampleDrms,
            bool useFullMap,
            int startRep,
            int repCount,
            int batch,
            RegressionStage stage)
        {
            if (!Config.DataSources.ContainsKey(dataSource))
            {
                throw new Exception(string.Format("Invalid data source: {0}\n Valid values:\n\t{1}", dataSource, string.Join("\n\t", Config.DataSources.Keys)));
            }

            Config config = new Config(xmlFile);

            if (startRep == default(int))
            {
                startRep = 0;
            }

            if (repCount == default(int))
            {
                repCount = 1;
            }

            string[] cellLines = config.CellLines;
            if (cellLineOverride != null)
            {
                cellLines = new string[] { cellLineOverride };
            }

            foreach (string cellLine in cellLines)
            {  
                Console.WriteLine("Loading drms for cell line {0}", cellLine);
                BedFile drmFile = new BedFile(config.GetCrmFileNameForCellLine(cellLine), BedFile.Bed4Crm);

                Console.WriteLine("Loading tss data for {0}", cellLine);
                List<Dictionary<string, BedFile.Location>> tssData = GetTssDataFromExpression(
                    config.RnaSources, 
                    i => config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources[i]), cellLine));

                Console.WriteLine("Loading NN map for cell line and tss data");

                List<ILookup<string, string>> filteredNNMaps = GetNNMapForVariantExpressedGenes(cellLine, config.RnaSources, drmFile, tssData, useFullMap);

                MultipleRegression(
                    config,
                    cellLine,
                    dataSource,
                    drmFile,
                    tssData,
                    filteredNNMaps,
                    generateFeatures,
                    sampleDrms,
                    useFullMap,
                    startRep,
                    repCount,
                    batch,
                    stage);
            }
        }

        private static Dictionary<string, BedFile.Location> UnionTssLocations(List<Dictionary<string, BedFile.Location>> tssData)
        {
            return tssData.Select(x => x)
                    .SelectMany(x => x)
                    .ToLookup(x => x.Key, x => x.Value)
                    .ToDictionary(x => x.Key, x => x.First());
        }

      /// <summary>
        /// Does regression across thresholds of a given cell line and set of expression data
        /// </summary>
        /// <param name="config">Config.</param>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="dataSource">Data source.</param>
        /// <param name="drmFile">Drm file.</param>
        /// <param name="tssData">Tss data.</param>
        /// <param name="variantExpressedNNMaps">Variant expressed NN maps.</param>
        /// <param name="generateFeatures">If set to <c>true</c> generate features.</param>
        /// <param name="sampleDrms">If set to <c>true</c> sample drms.</param>
        /// <param name="useFullMap">If set to <c>true</c> then do not leave the source cell line out of map generation</param>
        /// <param name="startRep">Start rep.</param>
        /// <param name="repCount">Rep count.</param>
        private static void MultipleRegression(
            Config config,
            string cellLine,
            string dataSource,
            BedFile drmFile,
            List<Dictionary<string, BedFile.Location>> tssData,
            List<ILookup<string, string>> variantExpressedNNMaps,
            bool generateFeatures,
            bool sampleDrms,
            bool useFullMap,
            int startRep,
            int repCount,
            int batch,
            RegressionStage stage)
        {
            int k = 0;
            int count = Utilities.GetChIPDataSetNames(config.GetChipDataSourceFileList(dataSource, cellLine)).Count();

            Dictionary<string, BedFile.Location> universalGeneIds = UnionTssLocations(tssData);

            switch (stage)
            {
                case RegressionStage.GenerateFeatures:
                {
                    HashSet<string> usedTsses = new HashSet<string>();
                    HashSet<string> usedDrms = new HashSet<string>();
                    for (int i = 0; i < config.SourceCount; i++)
                    {
                        for (int j = 0; j < config.Thresholds.Length; j++)
                        {
                            GC.Collect();
                            
                            //Console.WriteLine("Processing map {0} on threshhold {1}", i, thresholds[j]);
                            
                            Dictionary<string, List<string>> drmMap;
                            Dictionary<string, List<string>> drmMapNN;
                            
                            LoadMaps(config.GetMapFileNames(cellLine)[i][j], tssData[i], drmFile, variantExpressedNNMaps[i], out drmMap, out drmMapNN);

                            usedTsses.UnionWith(drmMap.Keys);
                            usedTsses.UnionWith(drmMapNN.Keys);
                            usedDrms.UnionWith(drmMap.Values.SelectMany(x => x));
                            usedDrms.UnionWith(drmMapNN.Values.SelectMany(x => x));
                        }

                        // Dictionary<string, int> drmIndices = 
                            TRFScorer.CreateIndex(drmFile.Locations, "DRM", Utilities.ScoreSource(cellLine, config.RnaSources[i]));
                        // Dictionary<string, int> tssIndices = 
                            TRFScorer.CreateIndex(universalGeneIds, "TSS", Utilities.ScoreSource(cellLine, config.RnaSources[i]));

                        foreach (string filename in Utilities.GetChIPDataFileNames(config.GetChipDataSourceFileList(dataSource, cellLine)))
                        {
                            var chipFile = new BedFile(filename, filename.Trim().Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout);

                            string baseFileName = Path.GetFileNameWithoutExtension(chipFile.FileName);
                            Console.WriteLine("\tScoring feature {2}: {0} of {1}", k, count, baseFileName);

                                TRFScorer.WriteScoreNameValuePairs(Utilities.GetDrmScoreFileName(baseFileName, cellLine, config.RnaSources[i]), TRFScorer.CreateDrmScores(drmFile, chipFile, false), null);
                                TRFScorer.WriteScoreNameValuePairs(Utilities.GetTssScoreFileName(baseFileName, cellLine, config.RnaSources[i]), TRFScorer.CreateTssScores(universalGeneIds, chipFile, config.D0, config.Range), null);

                            GC.Collect();
                            k++;
                        }
                    }

                    return;
               }
            }

            Func<string, string> MakeMapPromOnlyScript = tableName => tableName.Replace(".tsv", ".PromOnly.R");
            Func<string, string> MakeMapScript = tableName => tableName.Replace(".tsv", ".TssAndDrm.R");
            Func<string, string> MakeMapNNScript = tableName => tableName.Replace(".tsv", ".TssAndNNDrm.R");
            Func<string, string> MakeNNIntersectScript = tableName => tableName.Replace(".tsv", ".NNI.R");
            Func<string, string> MakeDifferenceScript = tableName => tableName.Replace(".tsv", ".D.R");
            Func<string, string> MakeDifferenceNNScript = tableName => tableName.Replace(".tsv", ".DNN.R");

            Func<string, string, string> MakeSampledPromOnlyScript = (tableName, theta) => tableName
                .Replace(".1NNR.", "." + theta + "NNR.")
                .Replace(".tsv", ".PromOnly.R");
            Func<string, string, string> MakeSampledNNScript = (tableName, theta) => tableName
                .Replace(".1NNR.", "." + theta + "NNR.")
                .Replace(".tsv", ".TssAndNNDrm.R");

            Func<string, string> MakeNNTableName = mapTableName => mapTableName.Replace(".tsv", "NN.tsv");

            Func<string, string> MakeMapNNIntersectTableName = mapTableName => mapTableName.Replace(".tsv", "NNI.tsv");
            Func<string, string> MakeDifferenceTableName = mapTableName => mapTableName.Replace(".tsv", "D.tsv");
            Func<string, string> MakeDifferenceNNTableName = mapTableName => mapTableName.Replace(".tsv", "DNN.tsv");

            // Create a map where the NN and correlation map agree
            Func<Dictionary<string, List<string>>, Dictionary<string, List<string>>, Dictionary<string, List<string>>> MapNNIntersect = 
                (drmMap, drmMapNN) =>
                drmMap.Select(x => new 
                    {
                        Tss = x.Key, 
                        CrmList = x.Value.Where(crm => drmMapNN [x.Key].Contains(crm))
                    })
                .Where(x => x.CrmList.Count() > 0)
                .ToDictionary(x => x.Tss, x => x.CrmList.ToList());

            // Create a map where the NN and correlation map disagree
            Func<Dictionary<string, List<string>>, Dictionary<string, List<string>>, Dictionary<string, List<string>>> MapDifference = 
                (drmMap, drmMapNN) =>
                drmMap.Select(x => new 
                    {
                        Tss = x.Key, 
                        CrmList = x.Value.Where(crm => drmMapNN [x.Key].Contains(crm)),
                        FullCrmList = x.Value
                    })
                .Where(x => x.CrmList.Count() == 0)
                .ToDictionary(x => x.Tss, x => x.FullCrmList);


            // Create a NN map where the map and NN disagree
            Func<Dictionary<string, List<string>>, Dictionary<string, List<string>>, Dictionary<string, List<string>>> MapDifferenceNN = 
                (drmMap, drmMapNN) => 
                drmMap.Select(x => new 
                    {
                        Tss = x.Key, 
                        CrmList = x.Value.Where(crm => drmMapNN [x.Key].Contains(crm)),
                        NNCrm = drmMapNN[x.Key]
                    })
                .Where(x => x.CrmList.Count() == 0)
                .ToDictionary(x => x.Tss, x => x.NNCrm);


            switch (stage)
            {
                case RegressionStage.BuildTables:
                    {


                        for (int i = 0; i < config.SourceCount; i++)
                        {
                            List<string> chipDataSets = Utilities.GetChIPDataSetNames(config.GetChipDataSourceFileList(dataSource, cellLine)).ToList();

                            var drmIndex = TRFScorer.ReadIndex("DRM", Utilities.ScoreSource(cellLine, config.RnaSources[i]));
                            var tssIndex = TRFScorer.ReadIndex("TSS", Utilities.ScoreSource(cellLine, config.RnaSources[i]));

                            List<double[]> drmScores = new List<double[]>();
                            List<double[]> tssScores = new List<double[]>();
                            foreach (string chipFile in chipDataSets)
                            {
                                Console.WriteLine("\tReading feature {2}: {0} of {1}", k++, chipDataSets.Count, Utilities.GetDrmScoreFileName(chipFile, cellLine, config.RnaSources[i]));
                                drmScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(drmIndex, Utilities.GetDrmScoreFileName(chipFile, cellLine, config.RnaSources[i]), x => double.Parse(x)));
                                tssScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(tssIndex, Utilities.GetTssScoreFileName(chipFile, cellLine, config.RnaSources[i]), x => double.Parse(x)));

                                GC.Collect();
                            }


                            string nnSampledMapOutFile = config.GetTableFileNames(cellLine)[i][0].Replace(".tsv", "NNR.tsv");

                            Dictionary<string, double> expression = GetExpression(config.RnaSources[i], config.GetExpressionFileName(config.RnaSources[i], cellLine));

                            //ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);
                            //Dictionary<string, string> ChIPToExpressionTSS_NN = NearestTSS(variantExpressedNNMaps[i].Select(x => x.Key), ExpressionTSSLocations);

                            var fullMapDict = variantExpressedNNMaps[i].Select(x => new { Tss = x.Key, Drms = new List<string>(x) } ).ToDictionary(x => x.Tss, x => x.Drms);
                            R.CreateTable(fullMapDict, tssIndex, tssScores, drmIndex, drmScores, expression, nnSampledMapOutFile, chipDataSets);

                            for (int j = 0; j < config.Thresholds.Length; j++)
                            {
                                Dictionary<string, List<string>> drmMap;
                                Dictionary<string, List<string>> drmMapNN;
                                //Dictionary<string, List<string>> drmMapAllDrms = new Dictionary<string, List<string>>();
                                
                                LoadMaps(config.GetMapFileNames(cellLine)[i][j], tssData[i], drmFile, variantExpressedNNMaps[i], out drmMap, out drmMapNN);

                                if (drmMap.Count >= tssScores.Count + drmScores.Count)
                                {
                                    string drmMapOutFile = config.GetTableFileNames(cellLine)[i][j];
                                    string nnMapOutFile = MakeNNTableName(drmMapOutFile);
                                    //string mapTssAllDrmOutFile = drmMapOutFile.Replace(".tsv", "DS.tsv");
                                    string mapNNIntersectFile = MakeMapNNIntersectTableName(drmMapOutFile);
                                    string mapDifferenceFile = MakeDifferenceTableName(drmMapOutFile);
                                    string mapDifferenceNNFile = MakeDifferenceNNTableName(drmMapOutFile);

                                    //Dictionary<string, string> ChIPToExpressionTSS = NearestTSS(drmMap.Select(x => x.Key), ExpressionTSSLocations);

                                    R.CreateTable(drmMap, tssIndex, tssScores, drmIndex, drmScores, expression, drmMapOutFile, chipDataSets);
                                    R.CreateTable(drmMapNN, tssIndex, tssScores, drmIndex, drmScores, expression, nnMapOutFile, chipDataSets);

                                    // Create a map where the NN and correlation map agree
                                    var mapNNIntersection = MapNNIntersect(drmMap, drmMapNN);
                                    R.CreateTable(mapNNIntersection, tssIndex, tssScores, drmIndex, drmScores, expression, mapNNIntersectFile, chipDataSets);

                                    // Create a map where the NN and correlation disagree
                                    var mapDifference = MapDifference(drmMap, drmMapNN);
                                    R.CreateTable(mapDifference, tssIndex, tssScores, drmIndex, drmScores, expression, mapDifferenceFile, chipDataSets);

                                    // Create a NN map where the map and NN disagree
                                    var mapDifferenceNN = MapDifferenceNN(drmMap, drmMapNN);
                                    R.CreateTable(mapDifferenceNN, tssIndex, tssScores, drmIndex, drmScores, expression, mapDifferenceNNFile, chipDataSets);

                                }
                            }
                        }
                        break;
                    }

                case RegressionStage.Regress:
                    {
                        for (int i = 0; i < config.SourceCount; i++)
                        {
                            List<string> chipDataSets = Utilities.GetChIPDataSetNames(config.GetChipDataSourceFileList(dataSource, cellLine)).ToList();

                            var tssColumns = R.MakeTssColumnNames(chipDataSets);
                            var drmColumns = R.MakeDrmColumnNames(chipDataSets);

                            Console.WriteLine("<table>");
                            Console.WriteLine(trhs + "Threshold" + th + "TssCount" +
                                //    th + "MapTssPromOnly" + //th + "MapTssPromOnlyStdErr" +
                                    th + "MapTssAndDrm" + //th + "MapTssAndDrmStdErr" +
                                    th + "MapTssAndNNDrm" + //th + "MapTssAndNNDrmStdErr" +
                                    th + "DiffCount" + th + "Pvalue" +
                                    th + "MapNNIntersectCount" + th +"NNIntersectionR2" + th + "NNSampleR2" + th + "NNIntersectionPvalue" +
                                    th + "MapNNDifferenceCount" + th + "DifferenceSqErr" + th + "NNDifferenceSqErr" + th + "DifferencePvalue" + 
                                              th + "SampledTssPromOnly" + th + "SampledTssAndNNDrm" +
                                //          th + "SampledTssPromOnlyDev" + th + "SampledTssAndNNDrmDev" + 
                                //              th + "SampledTssImprovementDev" +
                                   trhe);
                            
                            string nnSampledMapOutFile = config.GetTableFileNames(cellLine)[i][0].Replace(".tsv", "NNR.tsv");

                            //Dictionary<string, double> expression = GetExpression(config.GetExpressionFileName(config.RnaSources[i], cellLine));
                            
                            //ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);
                            //Dictionary<string, string> ChIPToExpressionTSS_NN = NearestTSS(variantExpressedNNMaps[i].Select(x => x.Key), ExpressionTSSLocations);
                            
                            //var fullMapDict = variantExpressedNNMaps[i].Select(x => new { Tss = x.Key, Drms = new List<string>(x) } ).ToDictionary(x => x.Tss, x => x.Drms);
                            //R.CreateTable(fullMapDict, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS_NN, nnSampledMapOutFile, chipDataSets);



                            for (int j = 0; j < config.Thresholds.Length; j++)
                            {
                                    Dictionary<string, List<string>> drmMap;// = LoadMap(config.GetMapFileNames(cellLine)[i][j], tssData[i])
                                    //.Select(x => new { Key = x.Key, Value = new List<string>(x) })
                                    //.ToDictionary(x => x.Key, x => x.Value);
                                    Dictionary<string, List<string>> drmMapNN;

                                LoadMaps(config.GetMapFileNames(cellLine)[i][j], tssData[i], drmFile, variantExpressedNNMaps[i], out drmMap, out drmMapNN);

                                // Create a map where the NN and correlation map agree
                                var mapNNIntersection = MapNNIntersect(drmMap, drmMapNN);

                                // Create a map where the NN and correlation disagree
                                var mapDifference = MapDifference(drmMap, drmMapNN);

                                // Create a NN map where the map and NN disagree
                                var mapDifferenceNN = MapDifferenceNN(drmMap, drmMapNN);

                                if (drmMap.Count >= tssColumns.Count() + drmColumns.Count())
                                {
                                    string drmMap_TableFile = config.GetTableFileNames(cellLine)[i][j];
                                    string nnMap_TableFile = MakeNNTableName(drmMap_TableFile);
                                    string mapNNIntersect_TableFile = MakeMapNNIntersectTableName(drmMap_TableFile);
                                    string mapDifference_TableFile = MakeDifferenceTableName(drmMap_TableFile);
                                    string mapDifferenceNN_TableFile = MakeDifferenceNNTableName(drmMap_TableFile);
                                    //string mapTss_AllDrmTableFile = drmMap_TableFile.Replace(".tsv", "DS.tsv");

                                    // Map TSS regression for promoter only and assigned DRMs

                                    string drmMap_ScriptFile_PromOnly = MakeMapPromOnlyScript(drmMap_TableFile);
                                    string drmMap_ScriptFile_TssDrm = MakeMapScript(drmMap_TableFile);

                                    double[] mapTss_Rsq_PromOnly = R.RegressColumns(
                                        drmMap_ScriptFile_PromOnly,
                                        drmMap_TableFile,
                                        tssColumns,
                                        R.RegressColumnsParam.AllRows,
                                        repCount,
                                        batch);

                                    double[] mapTss_Rsq_Drm = R.RegressColumns(
                                        drmMap_ScriptFile_TssDrm,
                                        drmMap_TableFile, 
                                        tssColumns.Concat(drmColumns),
                                        R.RegressColumnsParam.AllRows,
                                        repCount,
                                        batch);

                                    // NN DRM Regression on Map TSSes

                                    string nnMap_ScriptFile_TssDrm = MakeMapNNScript(nnMap_TableFile); 

                                    double[] mapTss_Rsq_NNDrm = R.RegressColumns(
                                        nnMap_ScriptFile_TssDrm,
                                        nnMap_TableFile,
                                        tssColumns.Concat(drmColumns),
                                        R.RegressColumnsParam.AllRows,
                                        repCount,
                                        batch);

                                    double[] mapNNIntersection_Rsq = new double[] { 0.0 };
                                    double[] mapNNIntersection_SampledControl_Rsq = new double[] { 0.0 };

                                    if (mapNNIntersection.Count >= tssColumns.Count() + drmColumns.Count())
                                    {
                                        string mapNNIntersect_ScriptFile = MakeNNIntersectScript(nnMap_TableFile);

                                        mapNNIntersection_Rsq = new double[] {R.RegressColumns(
                                            mapNNIntersect_ScriptFile,
                                            mapNNIntersect_TableFile,
                                            tssColumns.Concat(drmColumns))};

                                        string nnIntersectSampled_ScriptFile_TssDrm = MakeSampledNNScript(nnSampledMapOutFile, config.Thresholds[j]).Replace(".R", ".NNI.R");

                                        mapNNIntersection_SampledControl_Rsq = R.RegressColumns(
                                            nnIntersectSampled_ScriptFile_TssDrm,
                                            nnSampledMapOutFile,
                                            tssColumns.Concat(drmColumns),
                                            mapNNIntersection.Count,
                                            repCount);
                                    }

                                    double[] mapDifference_Rsq = new double[] { 0.0 };
                                    double[] mapDifferenceNN_Rsq = new double[] { 0.0 };

                                    if (mapDifference.Count >= tssColumns.Count() + drmColumns.Count())
                                    {
                                        string mapDifference_ScriptFile = MakeDifferenceScript(drmMap_TableFile);

                                        mapDifference_Rsq = R.RegressColumns(
                                            mapDifference_ScriptFile,
                                            mapDifference_TableFile,
                                            tssColumns.Concat(drmColumns),
                                            R.RegressColumnsParam.AllRows,
                                            repCount,
                                            batch);

                                        string mapDifferenceNN_ScriptFile = MakeDifferenceNNScript(nnMap_TableFile);

                                        mapDifferenceNN_Rsq = R.RegressColumns(
                                            mapDifferenceNN_ScriptFile,
                                            mapDifferenceNN_TableFile,
                                            tssColumns.Concat(drmColumns),
                                            R.RegressColumnsParam.AllRows,
                                            repCount,
                                            batch);
                                    }


                                    // Sampled tss prom only and nn regression

                                    string nnSampled_ScriptFile_PromOnly = MakeSampledPromOnlyScript(nnSampledMapOutFile, config.Thresholds[j]);
                                        
                                    double[] sampledTss_Rsq_PromOnly = R.RegressColumns(
                                        nnSampled_ScriptFile_PromOnly,
                                        nnSampledMapOutFile,
                                        tssColumns,
                                        drmMap.Count,
                                        repCount,
                                        batch);
                                                                    
                                    string nnSampled_ScriptFile_TssDrm = MakeSampledNNScript(nnSampledMapOutFile, config.Thresholds[j]);

                                    double[] sampledTss_Rsq_NNDrm = R.RegressColumns(
                                        nnSampled_ScriptFile_TssDrm,
                                        nnSampledMapOutFile,
                                        tssColumns.Concat(drmColumns),
                                        drmMap.Count,
                                        repCount,
                                        batch);

                                    int lessThanCount = mapTss_Rsq_Drm.Zip(mapTss_Rsq_NNDrm, (x, y) => x < y ? 1 : 0).Sum();
                                    double mapNNIntersectR2 = mapNNIntersection_Rsq.First();
                                    double mapNNIntersectPvalue = (double)Math.Min(mapNNIntersection_Rsq.Count(x => x < mapNNIntersectR2), mapNNIntersection_Rsq.Count(x => x > mapNNIntersectR2))
                                                         / (double)mapNNIntersection_SampledControl_Rsq.Count();

                                    int mapConflictLTCount = mapDifference_Rsq.Zip(mapDifferenceNN_Rsq, (x, y) => x < y ? 1 : 0).Sum();

                                    // Write R2 values for the batch
                                    //Console.WriteLine("Ranges " + mapTss_Rsq_Drm.Min() + ", " + mapTss_Rsq_Drm.Max());
                                    Console.WriteLine(trds + config.Thresholds[j] + td + drmMap.Count +

                                        //td + mapTss_Rsq_PromOnly.Average() + 
                                        td + mapTss_Rsq_Drm.Average() + 
                                        td + mapTss_Rsq_NNDrm.Average() + 
                                            //td + Math.Sqrt(mapTss_Rsq_PromOnly.Sum() / mapTss_Rsq_PromOnly.Length) + td + Stats.StdErr(mapTss_Rsq_PromOnly) +
                                            //          td + Math.Sqrt(mapTss_Rsq_Drm.Sum() / mapTss_Rsq_Drm.Length) + td + Stats.StdErr(mapTss_Rsq_Drm) +
                                            //          td + Math.Sqrt(mapTss_Rsq_NNDrm.Sum() / mapTss_Rsq_NNDrm.Length) + td + Stats.StdErr(mapTss_Rsq_NNDrm) +
                                            td + lessThanCount + " of " + mapTss_Rsq_Drm.Count() + 
                                            td + R.Binom(lessThanCount, mapTss_Rsq_Drm.Count()) +

                                        td + mapNNIntersection.Count + td + mapNNIntersectR2 + td + mapNNIntersection_SampledControl_Rsq.Average() + td + mapNNIntersectPvalue +
                                        td + mapDifference.Count + td +  mapDifference_Rsq.Average() + td + mapDifferenceNN_Rsq.Average() + td + R.Binom(mapConflictLTCount, mapDifferenceNN_Rsq.Count()) + 


                                                      td + sampledTss_Rsq_PromOnly.Average() + td + sampledTss_Rsq_NNDrm.Average() + 
                                        //              td + Stats.StdDev(sampledTss_Rsq_PromOnly) + td + Stats.StdDev(sampledTss_Rsq_NNDrm) + 
                                        //              td + Stats.StdDev(sampledTss_Rsq_NNDrm.Zip(sampledTss_Rsq_PromOnly, (x, y) => x - y)) +
                                         trde);
                                }
                            }

                            Console.WriteLine("</table>");
                        }
                        break;
                    }

                case RegressionStage.SampleDrms:
                {
                    //HashSet<string> drmWithVariedMarker;
                    //using (TextReader tr = new StreamReader(string.Format("Maps/InHouse/Correlation/{0}.All.Drm.{1}.nsv", cellLine, config.RnaSources[i])))
                    //{
                    //    drmWithVariedMarker = new HashSet<string>(tr.ReadToEnd().Split('\n'));
                    //}

                    //if (sampleDrms)
                    //{
                    //   drmMapAllDrms = drmMap.Where(mapEntry => tssData[i].ContainsKey(mapEntry.Key))
                    //        .Select(mapEntry => new 
                    //                {
                    //            Tss = mapEntry.Key,
                    //            Drms = GetDrmsInTssRange(tssData[i][mapEntry.Key], drmFile, 1000000, drmWithVariedMarker)
                    //        })
                    //            .ToDictionary(x => x.Tss, x => x.Drms.Select(y => y.Name).ToList());
                    //}
                    break;
                }

                case RegressionStage.Test:
                {
                    Console.WriteLine("<h1>" + cellLine + "</h1>");
                    for (int i = 0; i < config.SourceCount; i++)
                    {
                        Console.WriteLine("<h2>" + config.RnaSources[i] + "</h2>\n<table>");
                        string nnSampledMapOutFile = config.GetTableFileNames(cellLine)[i][0].Replace(".tsv", "NNR.tsv");

                        Console.WriteLine(trhs + "Theta" + th + "Map vs Prom" + th + "Map vs NNMap" + th + "NNMap vs Prom" + th + 
                                          "SNNMap vs SProm" + th + "Map vs SNNMap" + th + "NNMap vs SNNMap" + trhe);

                        for (int j = 0; j < config.Thresholds.Length; j++)
                        {
                            string drmMap_TableFile = config.GetTableFileNames(cellLine)[i][j];
                            string nnMap_TableFile = MakeNNTableName(drmMap_TableFile);

                            if (File.Exists(drmMap_TableFile))
                            {
                                string drmMap_ScriptFile_PromOnly = MakeMapPromOnlyScript(drmMap_TableFile);
                                string drmMap_ScriptFile_TssDrm = MakeMapScript(drmMap_TableFile);
                                string nnMap_ScriptFile_TssDrm = MakeMapNNScript(nnMap_TableFile); 
                                string nnSampled_ScriptFile_PromOnly = MakeSampledPromOnlyScript(nnSampledMapOutFile, config.Thresholds[j]);
                                string nnSampled_ScriptFile_TssDrm = MakeSampledPromOnlyScript(nnSampledMapOutFile, config.Thresholds[j]);

                                var mapPromOnlyR2 = R.ReadR2Set(drmMap_ScriptFile_PromOnly, batch);
                                var mapR2 = R.ReadR2Set(drmMap_ScriptFile_TssDrm, batch);
                                var nnMapR2 = R.ReadR2Set(nnMap_ScriptFile_TssDrm, batch);
                                var sampledNNPromOnlyR2 = R.ReadR2Set(nnSampled_ScriptFile_PromOnly, batch);
                                var sampledNNR2 = R.ReadR2Set(nnSampled_ScriptFile_TssDrm, batch);

                                //repCount = 100;
                                var pvalMap = Stats.StatisticalTests.PermutationTest(mapR2, mapPromOnlyR2, repCount);
                                var pvalNNMap = Stats.StatisticalTests.PermutationTest(nnMapR2, mapPromOnlyR2, repCount);
                                var pvalMapNNMap = Stats.StatisticalTests.PermutationTest(mapR2, nnMapR2, repCount);
                                var pvalSampledNN = Stats.StatisticalTests.PermutationTest(sampledNNR2, sampledNNPromOnlyR2, repCount);
                                var pvalMapSampledNN = Stats.StatisticalTests.PermutationTest(mapR2, sampledNNR2, repCount);
                                var pvalNNMapSampledNN = Stats.StatisticalTests.PermutationTest(nnMapR2, sampledNNR2, repCount);

                                Console.WriteLine(trds + config.Thresholds[j] + td + pvalMap + td + pvalMapNNMap + td + pvalNNMap + td + 
                                                  pvalSampledNN + td + pvalMapSampledNN + td + pvalNNMapSampledNN + trde);

                            }
                        }
                        Console.WriteLine("</table>");
                    }
                    break;
                }
            }









            /*for (int i = 0; i < config.SourceCount; i++)
            {
                Dictionary<string, double> expression = GetExpression(config.GetExpressionFileName(config.RnaSources[i], cellLine));

                Console.WriteLine("Make expression tss location lookup");
                ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);

                Dictionary<string, string> ChIPToExpressionTSS_NN = NearestTSS(variantExpressedNNMaps[i].Select(x => x.Key), ExpressionTSSLocations);

                HashSet<string> drmWithVariedMarker;
                
                using (TextReader tr = new StreamReader(string.Format("Maps/InHouse/Correlation/{0}.All.Drm.{1}.nsv", cellLine, config.RnaSources[i])))
                {
                    drmWithVariedMarker = new HashSet<string>(tr.ReadToEnd().Split('\n'));
                }

                var tssColumns = R.MakeTssColumnNames(chipDataSets);
                var drmColumns = R.MakeDrmColumnNames(chipDataSets);

                Console.WriteLine(trhs + "Threshold" + th + "TssCount" +
                                  th + "MapTssPromOnly" + th + "MapTssAndDrm" + th + "MapTssAndNNDrm" + 
                                  th + "SampledTssPromOnly" + th + "SampledTssAndNNDrm" + trhe);

                string nnSampledMapOutFile = config.GetOutputFileNames(cellLine)[i][0].Replace(".tsv", "NNR.tsv");

                var fullMapDict = variantExpressedNNMaps[i].Select(x => new { Tss = x.Key, Drms = new List<string>(x) } ).ToDictionary(x => x.Tss, x => x.Drms);
                R.CreateTable(fullMapDict, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS_NN, nnSampledMapOutFile, chipDataSets);

                for (int j = 0; j < config.Thresholds.Length; j++)
                {
                    GC.Collect();

                    Console.WriteLine("Processing map {0} on threshhold {1}", i, config.Thresholds[j]);
                    Console.Error.WriteLine("Processing map {0} on threshhold {1}", i, config.Thresholds[j]);

                    Dictionary<string, List<string>> drmMap;
                    Dictionary<string, List<string>> drmMapNN;
                    Dictionary<string, List<string>> drmMapAllDrms = new Dictionary<string, List<string>>();

                    LoadMaps(config.GetMapFileNames(cellLine)[i][j], tssData[i], drmFile, variantExpressedNNMaps[i], out drmMap, out drmMapNN);

                    if (sampleDrms)
                    {
                        drmMapAllDrms = drmMap.Where(mapEntry => tssData[i].ContainsKey(mapEntry.Key))
                            .Select(mapEntry => new 
                            {
                                Tss = mapEntry.Key,
                                Drms = GetDrmsInTssRange(tssData[i][mapEntry.Key], drmFile, 1000000, drmWithVariedMarker)
                            })
                            .ToDictionary(x => x.Tss, x => x.Drms.Select(y => y.Name).ToList());
                    }
                                                                                                                                                        ;

                    Dictionary<string, string> ChIPToExpressionTSS = NearestTSS(drmMap.Select(x => x.Key), ExpressionTSSLocations);

                    string drmMapOutFile = config.GetOutputFileNames(cellLine)[i][j];
                    string nnMapOutFile = drmMapOutFile.Replace(".tsv", "NN.tsv");
                    string mapTssAllDrmOutFile = drmMapOutFile.Replace(".tsv", "DS.tsv");

                    if (drmMap.Count >= tssScores.Count + drmScores.Count)
                    {
                        nTss[i][j] = R.CreateTable(drmMap, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, drmMapOutFile, chipDataSets);

                        double[] mapTssPromOnly = R.RegressColumns(drmMapOutFile, tssColumns, -1, repCount, drmMapOutFile.Replace(".tsv", ".PromOnly.R"));
                        double[] mapTssAndDrm = R.RegressColumns(drmMapOutFile, tssColumns.Concat(drmColumns), -1, repCount, drmMapOutFile.Replace(".tsv", ".TssAndDrm.R"));

                        R.CreateTable(drmMapNN, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, nnMapOutFile, chipDataSets);

                        double[] mapTssAndNNDrm = R.RegressColumns(nnMapOutFile, tssColumns.Concat(drmColumns), -1, repCount, nnMapOutFile.Replace(".tsv", ".TssAndNNDrm.R"));

                        double[] sampledTssPromOnly = R.RegressColumns(nnSampledMapOutFile, tssColumns, nTss[i][j], repCount, nnSampledMapOutFile.Replace(".tsv", ".PromOnly.R"));
                        double[] sampledTssAndNNDrm = R.RegressColumns(nnSampledMapOutFile, tssColumns.Concat(drmColumns), nTss[i][j], repCount, nnSampledMapOutFile.Replace(".tsv", ".TssAndNNDrm.R"));

                        //List<string> batchFiles = new List<string>();
                        //string[] outFileNames = { drmMapOutFile, nnMapOutFile, nnSampledMapOutFile, mapTssAllDrmOutFile };

                        /*if (sampleDrms)
                        {
                            batchFiles = R.CreateTableAndList(
                                drmMap, drmMapAllDrms, 
                                tssIndex, tssScores, 
                                drmIndex, drmScores, 
                                expression, ChIPToExpressionTSS, 
                                mapTssAllDrmOutFile, chipDataSets, repCount);
                        }*




                        //double[] yipMapScores =

                        //double[] Rsquareds;
                        //double[][] pvals;
                        //string scriptBaseName = drmMapOutFile.Replace(".tsv", ".R");
                        //R.BootRegressMapScoreData(outFileNames, batchFiles, chipDataSets, repCount, scriptBaseName, out Rsquareds, out pvals);

                        //allRsquareds[i][j] = Rsquareds;
                        //allPvals[i][j] = pvals;
                        //Console.Error.WriteLine("\t" + Rsquareds[0] + "\t" + Rsquareds[3] + "\t" + Rsquareds[6] + "\t" + Rsquareds[9]);

                        //double[] nnMapScores = R.RegressMapScoreData(nnMapOutFile, chipDataSets, R.RegressionFunction.Linear);

                        //double[] nnRandomMapScores = R.RegressMapScoreData(nnMapOutFile, chipDataSets, R.RegressionFunction.Linear);


                        // Bootstrap sampling from within C#
                        /*List<string> unsampledTssNames = new List<string>(tssNames);
                        var sampledTssNames = drmMap.Select(x => 
                        {
                            int sample = rng.Next(unsampledTssNames.Count);
                            string tss = unsampledTssNames[sample];
                            unsampledTssNames.RemoveAt(sample);
                            return tss;
                        });

                        Dictionary<string, List<string>> drmRandomNNTsses = sampledTssNames
                            .Select(x => new { Tss = x, Drms = fullNNMaps[i][x].ToList() })
                            .ToDictionary(x => x.Tss, x => x.Drms);*/


                        /*Console.WriteLine("NN random scores:\t{0}", string.Join("\t", nnRandomMapScores));
                   
                        allYipScores[i][j] = yipMapScores;
                        allNNScores[i][j] = nnMapScores;*
                    }
                    else
                    {
                        Console.Error.WriteLine("Skipping threshold {0} on map {1} due to insufficient map size", config.Thresholds[j], config.RnaSources[i]);  
                        double[] Rsquareds = new double[10].Select(x => double.NaN).ToArray();
                        double[][] pvals = new double[10].Select(x => new double[10].Select(y => double.NaN).ToArray()).ToArray();

                        allRsquareds[i][j] = Rsquareds;
                        allPvals[i][j] = pvals;
                    }
                }
            }

            /*string[] dataSets = { "Map", "NN", "RandomNN", "Map TSS/Sampled DRM" };
            string[] subset = { "DRM+TSS", "TSS", "DRM" };
            
            Func<int, string> DataSet = (index) => dataSets[index / 3];
            Func<int, string> DataSubset = (index) => subset[index % 3];

            int[] indices = { 0, 3, 6, 2, 5, 8, 1, 7, 9 };
            /// int[] DRMTSSindices = { 0, 3, 6 };
            /// int[] DRMindices = { 2, 5, 8 };
            /// int[] TSSindices = { 1, 7 };

            List<string> dataSetFullNames = indices.Select(x => string.Format("{0} {1}", DataSubset(x), DataSet(x))).ToList();

            for (int i = 0; i < config.SourceCount; i++)
            {
                Console.WriteLine("<table>");
                Console.WriteLine("<tr><th>-Log10 Threshhold</th><th colspan=10 align='center'>{0} Expression in {1}</th></tr>", config.RnaSources[i], cellLine);

                Console.WriteLine("<tr><th></th><th>TSS Count</th><th>" + string.Join("</th><th>", dataSetFullNames) + "</th></tr>");

                for (int j = 0; j < config.Thresholds.Length; j++)
                {
                    Console.WriteLine("<tr><td>{0}</td>", config.Thresholds[j]);
                    Console.WriteLine(
                        "<td>{0}</td><td>{1}</td>", 
                        nTss[i][j],
                        string.Join("</td><td>", indices.Select(x => string.Format("{0:F3}", allRsquareds[i][j][x])))
                        );
                    Console.WriteLine("</tr>");
                }
                Console.WriteLine("</table>");

            }

            for (int i = 0; i < config.SourceCount; i++)
            {
                Console.WriteLine("\n\n\n<table>");
                for (int j = 0; j < config.Thresholds.Length; j++)
                {
                    Console.WriteLine("<tr><td><h4>{0} {1} threshold {2}</h4>", config.RnaSources[i], cellLine, config.Thresholds[j]);
                    Console.WriteLine("<table>");

                    Console.WriteLine("<tr><td></td><td>{0}</td></tr>", string.Join("</td><td>", dataSetFullNames));

                    for (int l = 0; l < indices.Length; l++)
                    {
                        Console.WriteLine("<tr><td>{0}</td><td>{1}</td></tr>",
                            dataSetFullNames[l],
                            string.Join("</td><td>", indices.Select(x => allPvals[i][j][indices[l]][x])));
                    }

                    Console.WriteLine("</td></tr></table>");
                }
                Console.WriteLine("</table>");

            }*/
        }



        /*private static HashSet<string> FilterTranscripts(string configFile, string omittedCellLine)
        {
            XDocument doc = XDocument.Load(configFile);
            
            //string histoneRoot = doc.Descendants("HistoneRoot").First().Value.ToString();
            //string drmFileName = doc.Descendants("DrmFileName").First().Value.ToString();
            //string cellLineDrmFileName = doc.Descendants("CellLineDrmFileNames").First().Value.ToString();
            //int minFeatureCount = string.IsNullOrEmpty(omittedCellLine) ? 7 : 6;
            //BedFile drmFile = new BedFile(string.IsNullOrEmpty(omittedCellLine) ? drmFileName : string.Format(cellLineDrmFileName, omittedCellLine), BedFile.Bed4Drm);

            string expressionRoot = doc.Descendants("ExpressionRoot").First().Value.ToString();
            string[] cellLines = doc.Descendants("CellLines").First().Value.ToString().Split(',')
                .Where(cl => string.IsNullOrEmpty(omittedCellLine) || string.Compare(cl, omittedCellLine) != 0).ToArray();
            
            string[] expressionTypes = doc.Descendants("ExpressionTypes").First().Value.ToString().Split(',');

            Console.WriteLine("Building drm-tss pairs ...");
            
            // Expression data by type and cell line
            Dictionary<string, Dictionary<string, string>> validCellLineExpression = cellLines
                .Select(cl => expressionTypes
                        .Select(et => new
                        {
                    CellLine = cl,
                    ExpType = et,
                    Filename = string.Format("{0}/{1}/{2}.tsv", expressionRoot, cl, et)
                })
                .Where(expFile => File.Exists(expFile.Filename)))
                .SelectMany(x => x)
                .ToLookup(x => x.ExpType, x => x)
                .ToDictionary(x => x.Key, x => x.ToDictionary(y => y.CellLine, y => y.Filename));
            

            string testGene = "ENSG00000073111_127323094";

            var rawExpData = validCellLineExpression.Select(expressionType => 
            {
                Console.WriteLine("Loading {0} expression data ...", expressionType.Key);
                return expressionType.Value.Select(cl => 
                {
                    using (TextReader tr = new StreamReader(cl.Value))
                    {
                        Console.WriteLine("\t\tReading file {0}", cl.Value);
                        return tr.ReadToEnd()
                            .Split('\n')
                                .Where(line => !string.IsNullOrWhiteSpace(line))
                                .Select(line => 
                                        {
                                    string[] fields = line.Split('\t');
                                    return new 
                                    {
                                        CellLine = cl.Key,
                                        Tss = fields [0],
                                        Value = double.Parse(fields [1]),
                                        Chromosome = fields[2]
                                    };
                                });
                    }
                })
                .SelectMany(x => x)
                .ToLookup(x => x.Tss, x => x)
                .Where(x => x.Count() > 0)
                .Select(x => new 
                {
                    Tss = x.Key,
                    ExpressionType = expressionType.Key,
                    Chromosome = x.First().Chromosome,
                    Position = int.Parse(x.Key.Split('_')[1]),
                    Expression = x.ToLookup(y => y.CellLine, y => y.Value)
                        .ToDictionary(y => y.Key, y => y.Max())
                });
            })
            .SelectMany(x => x).ToList();

            Dictionary<string, int> tssWithMinExpAcrossAllTypes = rawExpData
                .Select(x => new {Tss = x.Tss, ExpList = x.Expression.Values })
                .Select(x => x.ExpList.Select(y => new { Tss = x.Tss, Exp = y }))
                .SelectMany(x => x)
                .ToLookup(x => x.Tss, x => x.Exp)
                .Select(x => new { Tss = x.Key, CountOver2 = x.Count(y => y >= 2) })
                .Where(x =>
                {
                        if (string.Compare(x.Tss, testGene) == 0)
                        {
                            Console.WriteLine("{1} count: {0}", x.CountOver2, testGene);
                        }
                        return x.CountOver2 > 0; 
                })
                .ToDictionary(x => x.Tss, x => x.CountOver2);

            int tssExpOver2 = tssWithMinExpAcrossAllTypes.Count();

            Console.WriteLine("Count of tss with expression over 2 across all cell lines and experiment types: {0}", tssExpOver2);

            Dictionary<string, int> tssWithMinVar = rawExpData
                .Where(x => tssWithMinExpAcrossAllTypes.ContainsKey(x.Tss))
                .Select(x => new {Tss = x.Tss, Experiment = x.ExpressionType, ExpList = x.Expression.Values })
                .Select(x => x.ExpList.Select(y => new { Tss = x.Tss, ET = x.Experiment, Exp = y }))
                .SelectMany(x => x)
                .ToLookup(x => x.Tss, x => x)
                .ToDictionary(x => x.Key, x => x.ToLookup(y => y.ET, y => y.Exp).Where(y => y.Count(z => z > 2) > 0))
                .Where(x => x.Value.Count() > 0)
                .Select(x =>
                {
                    if (string.Compare(x.Key, testGene) == 0)
                    {
                        foreach (var v in x.Value)
                        {
                                double m = v.Average();
                                Console.WriteLine(
                                    "{4} is two-fold: mean: {0}, max {1}, min {2}, var {5}: {3}", 
                                    v.Average(),
                                    v.Max(),
                                    v.Min(),
                                    IsTwoFoldChange(v),
                                    testGene,
                                    Math.Sqrt(v.Sum(a => Math.Pow(a - m, 2))));
                        }
                        
                    }
                    return new { Tss = x.Key, CountFoldChange = x.Value.Count(exptype => IsTwoFoldChange(exptype)) };
                })
                .Where(x => x.CountFoldChange > 0)
                .ToDictionary(x => x.Tss, x => x.CountFoldChange);

            int tssFolChangeOver2 = tssWithMinVar.Count();

            Console.WriteLine("{1} has count of fold change: {0}", tssWithMinVar[testGene], testGene);
            
            Console.WriteLine("Count of tss 2-fold change expression within each experiment type accross cell lines with only tss of min expression 2 somewhere: {0}", tssFolChangeOver2);



            //Console.WriteLine(string.Join("\n", tssWithMinVar.Keys));

            return new HashSet<string>(tssWithMinVar.Keys);
        }*/

       /// <summary>
        /// Create full bi-directional DRM and TSS centric nearest neighbor maps
        /// </summary>
        /// <returns>The drm NN map.</returns>
        /// <param name="tssData">Tss data.</param>
        /// <param name="drmFile">Drm file.</param>
        static ILookup<string, string> CreateDrmNNMap(
            Dictionary<string, BedFile.Location> tssData,
            BedFile drmFile,
            bool forwardOnly)
        {
            throw new Exception("Using bad NN mapping");

            Dictionary<string, List<BedFile.Location>> orderedTssStarts = GetOrderedTssStarts(tssData);

            Dictionary<string, List<BedFile.Location>> orderedDrmStarts = drmFile.GetOrderedLocations(x => x.Mid);

            //IEnumerable<Tuple<string, string>> drmNNmap = CreateDrmCentricNNMap(orderedTssStarts, orderedDrmStarts, drmFile.Locations);

            //IEnumerable<Tuple<string, string>> tssNNmap = CreateTssCentricNNMap(orderedTssStarts, orderedDrmStarts);

            if (forwardOnly)
            {
                return CreateTssCentricNNMap(orderedTssStarts, RemoveDrmsInPromoters(drmFile, orderedTssStarts)).ToLookup(x => x.Item1, x => x.Item2);
            } 
            else
            {
                return CreateDrmCentricNNMap(orderedTssStarts, orderedDrmStarts, drmFile.Locations)
                    .Concat(CreateTssCentricNNMap(orderedTssStarts, orderedDrmStarts))
                    .Distinct()
                    .ToLookup(x => x.Item1, x => x.Item2);
            }
        }

                    case "ScoreTssDrmNNPeaks":
                        ScoreTssDrmNNPeaks(
                            a.StringArgs ["TssFileName"], 
                            a.StringArgs ["DrmFileName"], 
                            a.StringArgs ["PeakFileName"], 
                            a.StringArgs ["ExpFileName"],
                            a.StringArgs ["OutputFileName"],
                            (double)a.IntArgs ["D0"],
                            a.IntArgs ["Range"]);
                        break;

static void ScoreTssDrmNNPeaks(
            string tssFileName,
            string drmFileName, 
            string peakFileName,
            string expFileName,
            string outputFileName,
            double D0,
            int range)
        {
            BedFile tssFile = new BedFile(tssFileName, BedFile.Bed4Crm);
            BedFile drmFile = new BedFile(drmFileName, BedFile.Bed4Crm);

            BedFile.Layout layout = peakFileName.Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout;
            BedFile peakFile = new BedFile(peakFileName, layout);

            var drmMap = CreateDrmNNMap(tssFile.Locations, drmFile, false);

            Console.WriteLine("Map created\n");

            var drmScores = TRFScorer.CreateDrmScores(drmFile, peakFile, false);

            var tssScores = TRFScorer.CreateTssScores(tssFile.Locations, peakFile, D0, range);

            Dictionary<string, double> expression = GetExpression("Unknown", expFileName);

            var v = from gene in expression
                join tss in tssScores on gene.Key equals tss.Key
                select new
            {
                Tss = gene.Key,
                TssOnlyScore = tss.Value,
                Score = (drmMap.Contains(tss.Key) ? drmMap [tss.Key].Sum(x => drmScores [x]) : 0.0),
                Exp = gene.Value
            };


            using (TextWriter tw = Helpers.CreateStreamWriter(outputFileName))
            { 
                tw.WriteLine("\"\",\"Gene\",\"TssScore\",\"TssDrmScore\",\"Expression\"");
                int i = 0;
                foreach (var tssScore in v.Where(x => drmMap.Contains(x.Tss)))
                {
                    tw.WriteLine("\"{0}\",{1},{2},{3},{4}", i, tssScore.Tss, tssScore.TssOnlyScore, tssScore.Score, tssScore.Exp);
                    i++;
                }
            }
        }



                    case "Regression":
                        RegressionDataUseXmlConfig(a.StringArgs ["Config"]);
                        break;

       private static void RegressionDataUseXmlConfig(string xmlFile)
        {
            XDocument doc = XDocument.Load(xmlFile);

            string tssFileName = doc.Descendants("TssFileName").First().Value.ToString();
            string chipFileName = doc.Descendants("ChIPFileListFileName").First().Value.ToString();
            string drmFileName = doc.Descendants("DrmFileName").First().Value.ToString();
            var mapData = doc.Descendants("MapFileName").FirstOrDefault();
            string mapFileName = mapData == null ? null : mapData.Value.ToString();
            string expFileName = doc.Descendants("ExpressionFileName").First().Value.ToString();
            string outFileName = doc.Descendants("OutputFileName").First().Value.ToString();
            double D0 = double.Parse(doc.Descendants("D0").First().Value.ToString());
            int range = int.Parse(doc.Descendants("PromoterRange").First().Value.ToString());

            GetRegressionData(
                tssFileName,
                chipFileName,
                drmFileName,
                mapFileName,
                expFileName,
                outFileName,
                D0,
                range);
        }

        private static void GetRegressionData(
            string tssFileName,
            string chipFileName,
            string drmFileName,
            string mapFileName,
            string expFileName,
            string outFileName,
            double D0,
            int range)
        {
            BedFile tssFile = new BedFile(tssFileName, BedFile.Bed4Crm);
            BedFile drmFile = new BedFile(drmFileName, BedFile.Bed4Crm);

            Dictionary<string, List<string>> drmMap;
            Dictionary<string, List<string>> drmMapNN;

            LoadMaps(mapFileName, tssFile.Locations, drmFile, CreateDrmNNMap(tssFile.Locations, drmFile, false), out drmMap, out drmMapNN);

            List<Dictionary<string, double>> drmScores = new List<Dictionary<string, double>>();
            List<Dictionary<string, double>> tssScores = new List<Dictionary<string, double>>();
            HashSet<string> geneIds = new HashSet<string>();

            List<string> chipDataSets = new List<string>();

            foreach (var chipFile in LoadChIPData(chipFileName))
            {
                drmScores.Add(TRFScorer.CreateDrmScores(drmFile, chipFile, false));
            
                var t = TRFScorer.CreateTssScores(tssFile.Locations, chipFile, D0, range);
                tssScores.Add(t);

                geneIds.UnionWith(tssScores.Last().Keys);
                chipDataSets.Add(chipFile.FileName);
            }

            Dictionary<string, double> expression = GetExpression("Unknown", expFileName);

            ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);

            Dictionary<string, string> ChIPToExpressionTSS = NearestTSS(drmMap.Select(x => x.Key), ExpressionTSSLocations);

            R.CreateTable(drmMap, tssScores, drmScores, expression, ChIPToExpressionTSS, outFileName, chipDataSets);

            Console.WriteLine(string.Join("\t", R.RegressMapScoreData(outFileName, chipDataSets, R.RegressionFunction.PrincipalComponents).Select(x => string.Format("{0}", x))));

            R.CreateTable(drmMapNN, tssScores, drmScores, expression, ChIPToExpressionTSS, outFileName.Replace(".tsv", "NN.tsv"), chipDataSets);
            
            Console.WriteLine(string.Join("\t", R.RegressMapScoreData(outFileName.Replace(".tsv", "NN.tsv"), chipDataSets, R.RegressionFunction.PrincipalComponents).Select(x => string.Format("{0}", x))));
        }

                    case "MapNearestGeneContent":   
                        MapContent(a.StringArgs ["MapFileName"], a.StringArgs ["GeneFileName"], a.StringArgs ["OutputFileName"], false);
                        break;

                    case "MapNearestDRMContent":
                        MapContent(a.StringArgs ["MapFileName"], a.StringArgs ["GeneFileName"], a.StringArgs ["OutputFileName"], true);
                        break;
                    
                    case "ScoreDrmPeaks":
                        ScoreDrmPeaks(a.StringArgs ["DrmFileName"], a.StringArgs ["PeakFileName"], a.StringArgs ["OutputFileName"]);
                        break;

                    case "ScoreTssPeaks":
                        ScoreTssPeaks(
                            a.StringArgs ["TssFileName"], 
                            a.StringArgs ["PeakFileName"], 
                            a.StringArgs ["OutputFileName"],
                            (double)a.IntArgs ["D0"],
                            a.IntArgs ["Range"]);
                        break;


                    /*case "FilterTranscripts":
                            Console.WriteLine(a.StringArgs.GetOptional("OmittedCellLine"));
                            FilterTranscripts(a.StringArgs["Config"], a.StringArgs.GetOptional("OmittedCellLine"));
                            break;*/

        static void MapContent(string mapFileName, string geneFileName, string outputFileName, bool fByDrm)
        {
            BedFile drms = new BedFile(mapFileName, BedFile.Bed4Crm);
            BedFile genes = new BedFile(geneFileName, BedFile.Bed4Crm);

            var outer = genes;
            var inner = drms;

            if (fByDrm)
            {
                outer = drms;
                inner = genes;
            }

            var pairs = 
               (from chromosome in outer.Chromosomes
                 where outer.ChromosomeBinnedLocations.ContainsKey(chromosome) &&
                       inner.ChromosomeBinnedLocations.ContainsKey(chromosome)
                select 
                       (from geneBin in outer.ChromosomeBinnedLocations[chromosome]
                        select new 
                        {
                            GeneBin = geneBin.Value, 
                            Drms = inner.ChromosomeBinnedLocations[chromosome]
                                         .Where(x => Math.Abs(x.Key - geneBin.Key) <= 1)
                                         .Select(x => x.Value)  // Get the name-location pairs from the bin
                                         .SelectMany(x => x)    // Flatten the bins' pairs into one enumerable
                        })
                        .Select(x => x.GeneBin  // Convert enumerable of bins to enumerable of genes
                                       .Select(g => 
                                        new 
                                        {
                                            Gene = g, 
                                            DrmAndDistance = x.Drms.Select(d => new {Drm = d, Distance = DrmGeneDistance(d.Value, g.Value)})
                                        }))
                        .SelectMany(x => x)
                 ).SelectMany(x => x)
                 .Where(x => x.DrmAndDistance.Count() > 0)
                 .Select(x => 
                     {
                        /*int min = int.MaxValue;
                        KeyValuePair<string, BedFile.Location> minDrm;
                        foreach (var v in x.DrmAndDistance)
                        {
                            if (v.Distance < min)
                            {
                                min = v.Distance;
                                minDrm = v.Drm;
                            }
                        }*/

                        var minDrmGene = x.DrmAndDistance.OrderBy(d => d.Distance).Where((d, i) => i < 1);
                        var key = new List<KeyValuePair<string, BedFile.Location>>(){x.Gene};
                        Dictionary<string, int> distances = new Dictionary<string, int>();
                        foreach (var drm in minDrmGene)
                        {
                            string geneName = fByDrm ? drm.Drm.Key : x.Gene.Key;
                            string drmName = fByDrm ? x.Gene.Key : drm.Drm.Key;
                            string edge = string.Format("{0},{1}", geneName, drmName);
                            distances.Add(edge, drm.Distance);
                        }

                        return new
                        {
                            // Swap gene and drm if using drm as the outer source
                            Genes = fByDrm ? minDrmGene.Select(d => d.Drm).ToList() : key,
                            Drms = fByDrm ? key : minDrmGene.Select(d => d.Drm).ToList(),
                            Distances = distances
                        };
                    }
                    );

            using (TextWriter tw = Helpers.CreateStreamWriter(outputFileName))
            {
                int n = 0;
                int count = pairs.Count();
                foreach (var x in pairs)
                {
                    string line = 
                        string.Join(
                            "\n", 
                                from gene in x.Genes
                                from drm in x.Drms
                                select 
                                    string.Format(
                                        "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}",
                                        drm.Value.Chromosome, 
                                        drm.Value.Start,
                                        drm.Value.End,
                                        drm.Value.Strand,
                                        gene.Key,
                                        drm.Key,
                                        x.Distances[string.Format("{0},{1}", gene.Key, drm.Key)]));
                    tw.WriteLine(line);
                    if (n % 10000 == 0)
                    {
                        Console.WriteLine("Processed {0} of {1} genes", n, count);
                    }
                    n++;
                }                 
            }
        }

        private static int DrmGeneDistance(BedFile.Location l1, BedFile.Location l2)
        {
            if (string.Compare(l1.Chromosome, l2.Chromosome) != 0)
            {
                return int.MaxValue;
            }

            return Math.Abs(l1.Mid - l2.Start);
        }

        static void ScoreDrmPeaks(string binFileName, string peakFileName, string outputFileName)
        {
            BedFile drmFile = new BedFile(binFileName, BedFile.Bed4Crm);

            BedFile.Layout layout = peakFileName.Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout;
            BedFile peakFile = new BedFile(peakFileName, layout);

            var drmScores = TRFScorer.CreateDrmScores(
                drmFile, 
                peakFile,
                false);

            foreach (var v in drmScores.Where(x => x.Value > 0))
            {
                    Console.WriteLine("{0}\t{1}", v.Key, v.Value);
            }
        }

        static void ScoreTssPeaks(string tssFileName, string peakFileName, string outputFileName, double D0, int maxRange)
        {
            throw new NotImplementedException();
            /*BedFile tssFile = new BedFile(tssFileName, BedFile.Bed4Drm);

            BedFile.Layout layout = peakFileName.Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout;
            BedFile peakFile = new BedFile(peakFileName, layout);

            var tssScores = CreateTssScores(tssFile, peakFile, D0, maxRange);
         
            foreach (var v in tssScores.Where(x => x.Value > 0))
            {
                Console.WriteLine("{0}\t{1}", v.Key, v.Value);
            }*/
        }


case "ProgressivePromoterAdditions":
ProgressivePromoterAdditions(
    a.StringArgs ["Config"],
    a.StringArgs ["CellLine"],
    a.StringArgs.GetOptional("UseVariantExpressedSubset") == "True");
break;

/// <summary>
/// Perform genome wide regression (or on variantly expressed subset) on progressively more complicated models
/// by adding TSS-centric NN, DRM-centric NN and K NN TSS-centric 
/// </summary>
/// <param name="xmlFile">Xml file.</param>
/// <param name="cellLine">Cell line.</param>
public static void ProgressivePromoterAdditions(string xmlFile, string cellLine, bool fUseVariantExpressedSubset)
{
    Config config = new Config(xmlFile);

    string chipFileName = config.CellLineTag.Replace(config.GetChipDataSourceFileList("AllChIP"), cellLine);

    List<string> chipDataSets = Utilities.GetChIPDataSetNames(chipFileName).ToList();


    var drmIndex = TRFScorer.ReadIndex("DRM", cellLine);
    var tssIndex = TRFScorer.ReadIndex("TSS", cellLine);

    List<double[]> drmScores = new List<double[]>();
    List<double[]> tssScores = new List<double[]>();

    int k = 0;
    foreach (string chipFile in chipDataSets)
    {
        Console.WriteLine("\tReading feature {2}: {0} of {1}", k++, chipDataSets.Count, Utilities.GetDrmScoreFileName(chipFile, cellLine, "NA"));
        drmScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(drmIndex, Utilities.GetDrmScoreFileName(chipFile, cellLine, "NA"), x => double.Parse(x)));
        tssScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(tssIndex, Utilities.GetTssScoreFileName(chipFile, cellLine, "NA"), x => double.Parse(x)));

        GC.Collect();
    }

    Console.WriteLine("Loading drms for cell line {0}", cellLine);
    BedFile drmFile = new BedFile(config.GetCrmFileNameForCellLine(cellLine), BedFile.Bed4Crm);

    Console.WriteLine("Loading tss data for {0}", cellLine);
    List<Dictionary<string, BedFile.Location>> tssData = GetTssDataFromExpression(
        config.RnaSources, 
        i => config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources[i]), cellLine));

    for (int i = 0; i < config.RnaSources.Length; i++)
    {   
        Dictionary<string, double> expression = GetExpression(config.RnaSources [i], config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources [i]), cellLine));

        Dictionary<string, BedFile.Location> drmFilteredLocations = RemoveDrmsInPromoters(drmFile, GetOrderedTssStarts(tssData[i]))
            .Select(chr => chr.Value.Select(drm => new { Key = drm.Name, Value = drm}))
            .SelectMany(x => x)
            .ToDictionary(x => x.Key, x => x.Value);

        HashSet<string> drmSet = fUseVariantExpressedSubset ? GetVariedElement("Drm", cellLine, config.RnaSources[i], false) : new HashSet<string>(drmFilteredLocations.Keys);
        Dictionary<string, BedFile.Location> drmLocations = drmFilteredLocations.Where(x => drmSet.Contains(x.Key)).ToDictionary(x => x.Key, x => x.Value);;
        var orderedDrmStarts = GetOrderedFilteredLocations(drmFilteredLocations, drmSet);

        HashSet<string> tssSet = fUseVariantExpressedSubset ? GetVariedElement("Tss", cellLine, config.RnaSources[i], false) : new HashSet<string>(tssData[i].Keys);
        var orderedFilteredTssStarts = GetOrderedFilteredLocations(tssData[i], tssSet);

        var tssNNMap = CreateTssCentricNNMap(orderedFilteredTssStarts, orderedDrmStarts)
            .Where(x => tssSet.Contains(x.Item1) && drmSet.Contains(x.Item2))
            .ToLookup(x => x.Item1, x => x.Item2)
            .ToDictionary(x => x.Key, x => x.ToList());

        var drmNNMap = CreateDrmCentricNNMap(orderedFilteredTssStarts, orderedDrmStarts, drmLocations)
            .Where(x => tssSet.Contains(x.Item1) && drmSet.Contains(x.Item2))
            .ToLookup(x => x.Item1, x => x.Item2)
            .ToDictionary(x => x.Key, x => x.ToList());

        ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);

        Dictionary<string, string> ChIPToExpressionTSS = NearestTSS(tssNNMap.Select(x => x.Key), ExpressionTSSLocations);

        Console.WriteLine("TSS NN Map has {0} TSSes", tssNNMap.Count());
        Console.WriteLine("DRM NN Map has {0} TSSes", drmNNMap.Count());

        string tableFile = "../temp/PPA." + cellLine + ".tsv";
        int nTss = R.CreateTable(tssNNMap, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, tableFile, chipDataSets);

        var tssColumns = R.MakeTssColumnNames(chipDataSets);
        var drmColumns = R.MakeDrmColumnNames(chipDataSets);

        int repCount = 1000;

        double[] allTss_PromOnly_RSquareds = R.RegressColumns("../temp/PPA/allTss_PO.R", tableFile, tssColumns, R.RegressColumnsParam.AllRows, repCount, 1);
        double[] allTss_TssNN_RSquareds = R.RegressColumns("../temp/PPA/allTss_NN.R", tableFile, tssColumns.Concat(drmColumns), R.RegressColumnsParam.AllRows, repCount, 1);

        Console.WriteLine("All tss count: " + nTss);
        Console.WriteLine("All tss prom only: {0}", allTss_PromOnly_RSquareds.Average());
        Console.WriteLine("All tss tss NN: {0}", allTss_TssNN_RSquareds.Average());

        var tssNNMapDrmNNSubset = tssNNMap.Where(x => drmNNMap.ContainsKey(x.Key)).ToDictionary(x => x.Key, x => x.Value);

        string dataFileDrmNN = "../temp/PPA2." + cellLine + ".tsv";
        int drmNNTssCount = R.CreateTable(drmNNMap, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, dataFileDrmNN, chipDataSets);

        string dataFileTssNNDrmSubset = "../temp/PPA3." + cellLine + ".tsv";
        R.CreateTable(tssNNMapDrmNNSubset, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, dataFileTssNNDrmSubset, chipDataSets);

        double[] subsetTss_PromOnly_RSquareds = R.RegressColumns("../temp/PPA/subsetTss_PO.R", dataFileDrmNN, tssColumns, R.RegressColumnsParam.AllRows, repCount, 1);
        double[] subsetTss_DrmNN_RSquareds = R.RegressColumns("../temp/PPA/subsetTss_DRMNN.R", dataFileDrmNN, tssColumns.Concat(drmColumns), R.RegressColumnsParam.AllRows, repCount, 1);
        double[] subsetTss_TssNN_RSquareds = R.RegressColumns("../temp/PPA/subsetTss_TSSNN.R", dataFileTssNNDrmSubset, tssColumns.Concat(drmColumns), R.RegressColumnsParam.AllRows, repCount, 1);



        //double pvalue;
        //double[] subsetTss_TssNN_ShuffledDRMs = R.RegressPromAndShuffledDrm(dataFile, chipDataSets, out pvalue);

        //Console.WriteLine("Shuffled DRMs R2s: prom: " + subsetTss_TssNN_ShuffledDRMs[0] + "; tss shuffled: " + subsetTss_TssNN_ShuffledDRMs[1] + "; pvalue = " + pvalue);

        Console.WriteLine("Subset tss count: " + drmNNTssCount);
        Console.WriteLine("Subset tss prom only: {0}", subsetTss_PromOnly_RSquareds.Average());
        Console.WriteLine("Subset tss drm NN: {0}", subsetTss_DrmNN_RSquareds.Average());
        Console.WriteLine("Subset tss tss NN: {0}", subsetTss_TssNN_RSquareds.Average());

        Console.WriteLine("<table>" + trhs + "TSS Set" + th + "TSS Count" + trhe);
        Console.WriteLine(trds + "All TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : "") + td + nTss + trde);
        Console.WriteLine(trds + "DRM NN TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : "") + td + drmNNTssCount + trde + "</table>");

        Console.WriteLine("<table>" + trhs + "Regression Model" + th + "Adjusted R2" + trhe);
        Console.WriteLine(trds + "All TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "Promoter Only" + td + allTss_PromOnly_RSquareds.Average() + trde);
        Console.WriteLine(trds + "All TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "TSS NN Map" + td + allTss_TssNN_RSquareds.Average() + trde);
        Console.WriteLine(trds + "DRM NN TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "Promoter Only" + td + subsetTss_PromOnly_RSquareds.Average() + trde);
        Console.WriteLine(trds + "DRM NN TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "DRM NN Map" + td + subsetTss_DrmNN_RSquareds.Average() + trde);
        Console.WriteLine(trds + "DRM NN TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "TSS NN Map" + td + subsetTss_TssNN_RSquareds.Average() + trde + "</table>");

        Console.WriteLine("<table>" + trhs + "Comparison" + th + "pvalue" + trhe);
        Console.WriteLine(trds + "All TSS Prom vs TSS NN Map" + td + Stats.StatisticalTests.PermutationTest(allTss_PromOnly_RSquareds, allTss_TssNN_RSquareds, 100000) + trde);
        Console.WriteLine(trds + "DRM NN TSS Prom vs DRM NN Map" + td + Stats.StatisticalTests.PermutationTest(subsetTss_PromOnly_RSquareds, subsetTss_DrmNN_RSquareds, 100000) + trde);
        Console.WriteLine(trds + "DRM NN TSS Prom vs TSS NN Map" + td + Stats.StatisticalTests.PermutationTest(subsetTss_PromOnly_RSquareds, subsetTss_TssNN_RSquareds, 100000) + trde);
        Console.WriteLine(trds + "DRM NN TSS DRM NN Map vs TSS NN Map" + td + Stats.StatisticalTests.PermutationTest(subsetTss_DrmNN_RSquareds, subsetTss_TssNN_RSquareds, 100000) + trde);
    }
}


case "GetTFGeneSets":
GetTFGeneSets(
    a.StringArgs ["Config"],
    a.StringArgs ["CellLine"]
);
break;

/// <summary>
/// Gets the set of genes 'targeted' by a TF for all TFs chipped in a cell line.
/// TF targets are determined by binding in the 10KB promoter region of any gene,
/// the 10KB promoter region of the genes in a map, or either the promoter region or 
/// a CRM of a gene in a map
/// </summary>
/// <param name="xmlFile">Xml file.</param>
/// <param name="cellLine">Cell line.</param>
private static void GetTFGeneSets(string xmlFile, string cellLine)
{
    Config config = new Config(xmlFile);

    Dictionary<string, BedFile.Location> tssData = GetExpressionTssData(
        config.RnaSources [0], 
        config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources [0]), cellLine));



    //Console.WriteLine("Processing map {0} on threshhold {1}", i, thresholds[j]);

    Dictionary<string, List<string>> drmMap = LoadMap(config.GetMapFileNames(cellLine) [0] [0], tssData)
        .Select(x => new { Key = x.Key, Value = new List<string>(x) })
        .ToDictionary(x => x.Key, x => x.Value);
    //Dictionary<string, List<string>> drmMapNN;


    Console.WriteLine("Loading drms for cell line {0}", cellLine);
    BedFile drmFile = new BedFile(config.GetCrmFileNameForCellLine(cellLine), BedFile.Bed4Crm);



    Console.WriteLine("Loading NN map for cell line and tss data");
    //List<ILookup<string, string>> filteredNNMaps = GetNNMapForVariantExpressedGenes(cellLine, config.RnaSources, drmFile, tssData, f, false);

    //LoadMaps(config.GetMapFileNames(cellLine)[0][0], tssData[0], drmFile, variantExpressedNNMaps[0], out drmMap, out drmMapNN);


    foreach (string filename in Utilities.GetChIPDataFileNames(config.GetChipDataSourceFileList("TRF", cellLine)))
    {
        var chipFile = new BedFile(filename, filename.Trim().Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout);

        string baseFileName = Path.GetFileNameWithoutExtension(chipFile.FileName);
        Console.WriteLine("\tGetting targets of TF {0}", baseFileName);


        var tfTargetData = tssData
            .Where(x => chipFile.ChromosomeOrderedLocations.ContainsKey(x.Value.Chromosome))
            .Select(x => 
                {
                    bool inMap = drmMap.ContainsKey(x.Key);
                    bool crmContainsPeak = false;

                    var orderedLocations = chipFile.ChromosomeOrderedLocations[x.Value.Chromosome];

                    if (inMap)
                    {
                        // Find any overlapping peak with any CRM
                        crmContainsPeak = drmMap[x.Key].Count(crm => TRFScorer.EvaluateRange(
                            orderedLocations, 
                            drmFile.Locations[crm].Start, 
                            drmFile.Locations[crm].End, 
                            0, 
                            (item, start) => item.End < start,
                            (item, end) => item.Start <= end,
                            (count, length, location) => count + 1) > 0) > 0;
                    }

                    return new 
                    {
                        Tss = x.Key,
                        Location = x.Value,
                        ClosestPeak = TRFScorer.GetClosestMid(orderedLocations, x.Value.DirectionalStart),
                        InMap = inMap,
                        CrmContainsPeak = crmContainsPeak
                    };
                });

        Func<BedFile.Location, BedFile.Location, bool> inPromoter = (tss, peak) =>
                                                                            Math.Abs(tss.DirectionalStart - peak.Start) <= 10000 || Math.Abs(tss.DirectionalStart - peak.End) <= 10000;

        if (!Directory.Exists("../temp/TFCRMTargets/"))
        {
            Directory.CreateDirectory("../temp/TFCRMTargets/");
        }

        using (TextWriter tw = Helpers.CreateStreamWriter("../temp/TFCRMTargets/AllProm." + cellLine + "." + baseFileName + ".nsv"))
        {
            tw.WriteLine(string.Join("\n", tfTargetData.Where(x => inPromoter(x.Location, x.ClosestPeak))
                .Select(x => x.Tss))); 
        }

        using (TextWriter tw = Helpers.CreateStreamWriter("../temp/TFCRMTargets/MapProm." + cellLine + "." + baseFileName + ".nsv"))
        {
            tw.WriteLine(string.Join("\n", tfTargetData.Where(x => inPromoter(x.Location, x.ClosestPeak) && x.InMap)
                .Select(x => x.Tss))); 
        }
        using (TextWriter tw = Helpers.CreateStreamWriter("../temp/TFCRMTargets/AllProm." + cellLine + "." + baseFileName + ".nsv"))
        {
            tw.WriteLine(string.Join("\n", tfTargetData.Where(x => inPromoter(x.Location, x.ClosestPeak) && x.InMap || x.CrmContainsPeak)
                .Select(x => x.Tss))); 
        }
    }
}